<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 6917b9add631 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-04-21 13:24 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_base_showFile{

	public static function openBrowser(int $id, string $url): void{
		session_write_close();
		if($id){
			$path = fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE IsPublished=1 AND ID=' . $id);
			if(!$path){
				echo g_l('global', '[site_not_published]');
				return;
			}
			$urlReplace = we_folder_base::getUrlReplacements(DB_WE::$inst, true);
			$url = ($urlReplace ?
				(($http = preg_replace($urlReplace, array_keys($urlReplace), $path, -1, $cnt)) && $cnt ? '' : we_http_util::getServerUrl()) . $http :
				we_http_util::getServerUrl() . $path
				) . '?r=' . microtime();
		}
		header('Location: ' . $url);
		echo we_html_tools::getHtmlTop('', '<meta http-equiv="refresh" content="1; url=' . $url . '">', we_html_element::body());
	}

	public static function showTempFile(string $filename, bool $keepBin): void{
		session_write_close();
		if(strpos($filename, realpath(WE_INCLUDES_PATH) ?: '') !== false){
			//nobody should read inside include directory
			return;
		}

		if(!file_exists($filename)){
			return;
		}
		$isCompressed = $keepBin ? false : we_base_file::isCompressed($filename);
		$allfile = explode('.', $filename);

		if(!$keepBin && ($mimetype = we_base_util::getMimeType(end($allfile) ?: '', $filename, we_base_util::MIME_BY_HEAD_THEN_EXTENSION, true))){
			switch($mimetype){
				case 'text/plain': //let the browser decide
				case 'application/x-empty':
					break;
				default:
					header('Content-Type: ' . $mimetype . '; charset=UTF-8');
			}
		} else {
			header('Content-Type: application/octet-stream');
			header('Content-Disposition: attachment; filename="' . ($keepBin ? basename($filename) : basename($filename, '.gz')) . '"');
			$isCompressed = false;
		}

		if($isCompressed){
			readgzfile($filename);
		} else {
			readfile($filename);
		}
	}

	public static function showThumbnail(int $imageId, int $thumbID): void{
		session_write_close();
		if($thumbID){
			$img = new we_document_image($imageId);
			$thumbObj = new we_contents_thumbnail();
			$thumbObj->initByThumbID($thumbID, $img->ID, $img->Filename, $img->Path, $img->Extension, $img->getElementI('origwidth'), $img->getElementI('origheight'), $img->getDocument());
			if(!$thumbObj->exists()){
				$thumbObj->getThumb();
			}
			$file = $thumbObj->getOutputPath(true, false);
			$imageExt = $thumbObj->getOutputformat();
		} else {
			$imagePath = (string) we_base_request::_(we_base_request::FILE, 'path', '');
			$imageSizeW = we_base_request::int('size', 0, 'width');
			$imageSizeH = we_base_request::int('size', $imageSizeW, 'height');
			$extension = we_base_request::string('extension', '');

			if(!($imageId || $imagePath) && !$imageSizeW && !$extension){
				exit();
			}

			$whiteList = we_base_ContentTypes::getExtension(we_base_ContentTypes::IMAGE);

			if(!in_array(strtolower($extension), $whiteList, true)){
				exit();
			}

			$imageExt = substr($extension, 1) ?: '';
			try{
				$file = we_base_imageEdit::createPreviewThumb($imagePath, $imageId, $imageSizeW, $imageSizeH, $imageExt);
			} catch (we_exception_base $e){
				t_e($e);
				we_html_tools::setHttpCode(404);
				return;
			}
		}
		if(!file_exists($file) || !is_readable($file)){
			we_html_tools::setHttpCode(404);
			return;
		}
		$stat = stat($file);
		$etag = md5_file($file);
		$modDate = gmdate('D, d M Y H:i:s', $stat['mtime']) . ' GMT';
		header('Etag: "' . $etag . '"');
		header('Expires: ' . gmdate('D, d M Y H:i:s', time() + 60) . ' GMT', true);
		header('Last-Modified: ' . $modDate, true);
		header('Cache-Control: max-age=60'); //they stay in cache for 60 seconds, before reasking the server for a new version!
		header_remove('Pragma');

		if((!empty($_SERVER['HTTP_IF_NONE_MATCH']) && $_SERVER['HTTP_IF_NONE_MATCH'] === $etag) ||
			(!empty($_SERVER['HTTP_IF_MODIFIED_SINCE']) && $_SERVER['HTTP_IF_MODIFIED_SINCE'] === $modDate)){
			we_html_tools::setHttpCode(304);
			return;
		}
		header('Content-type: image/' . $imageExt);
		header('Content-Length: ' . filesize($file));
		readfile($file);
	}

	/**
	 *
	 * @param string $uniqid
	 * @param string $we_transaction
	 * @param int[] $thumbIDs
	 * @return void
	 */
	public static function showThumbTable(string $uniqid, string $we_transaction, array $thumbIDs): void{
		session_write_close();
		if(!($uniqid && $we_transaction && $thumbIDs)){
			we_html_tools::setHttpCode(404);
			return;
		}
		$we_doc = we_contents_root::getSessionDocument($we_transaction);
		assert($we_doc instanceof we_document_image);

		$table = '<table class="default"><tr>';

		foreach($thumbIDs as $thumbid){
			$thumbObj = new we_contents_thumbnail();
			$thumbObj->initByThumbID($thumbid, $we_doc->ID, $we_doc->Filename, $we_doc->Path, $we_doc->Extension, $we_doc->getElementI('origwidth'), $we_doc->getElementI('origheight'), $we_doc->getDocument());

			$randval = microtime();

			$useOrig = $thumbObj->isOriginal();

			if((!$useOrig) && $we_doc->ID && ($we_doc->DocChanged === false) && file_exists($thumbObj->getOutputPath(true))){
				$src = $thumbObj->getOutputPath(false, true);
			} else {
				$src = WEBEDITION_DIR . 'we_cmd.php?' . http_build_query([
						'we_cmd' => [
							0 => 'we_base_showBinaryDoc',
							1 => $we_doc->ContentType,
							2 => $we_transaction,
							3 => ($useOrig ? '' : $thumbid),
						],
						'rand' => $randval
				]);
			}

			$table .= '<td><image src="' . $src . '" style="width:' . $thumbObj->getOutputWidth() . 'px;height:' . $thumbObj->getOutputHeight() . 'px"/></td>';
		}

		$table .= '</tr></table>';
		echo we_html_tools::getHtmlTop('', '', we_html_element::body(['style' => 'margin: 5px;'], $table));
	}

}
