<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 2086f1f2c8de $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-04-06 18:57 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_captcha_image{
	/** @var int Length of the text	 */
	private $textlength = 0;

	/** @var int Width of the image	 */
	private $width = 0;

	/** @var int Height of the image	 */
	private $height = 0;

	/** @var int[] Backgroundcolor of the image	 */
	private $background = [0, 0, 0];

	private const DEFAULT_FONT = [
		'size' => [
			'min' => 15,
			'max' => 20
		],
		'color' => []
	];

	/** @var array{size:array{min:int,max:int},color:int[][]} all informations about the used font	 */
	private $font = self::DEFAULT_FONT;

	/** @var string  path where font is located	 */
	private $fontfile = WE_LIB_PATH . 'fonts/DejaVuSans.ttf';

	/** @var array range of the angle	 */
	private $angle = [];

	/** @var string vertical align of the code	 */
	private $valign = '';

	/** @var string align of the code */
	private $align = '';

	private const DEFAULT_STYLE = [
		'strikeout' => false,
		'fullcircle' => false,
		'outlinecircle' => false,
		'fullrectangle' => false,
		'outlinerectangle' => false,
		'color' => [],
		'number' => [
			'min' => 0,
			'max' => 0
		],
	];

	/** @var array strikeout */
	private $style = self::DEFAULT_STYLE;

	/** @var array ranges of the charactersubset */
	private $charactersubset = [];

	/** @var bool transparent */
	private $transparent = false;

	/** @var int[] */
	private $skip = [];

	/**
	 * PHP5 Constructor
	 *
	 * @param integer $width
	 * @param integer $height
	 * @param integer $textlength
	 * @return void
	 */
	function __construct(int $width, int $height, int $textlength = 5){
		$this->width = $width;
		$this->height = $height;

		$this->textlength = max(3, min(32, $textlength));

		// init the font
		$this->setFont();

		// init the character subset
		$this->setCharacterSubset();

		// no signs to skip
		$this->skip = [];

		// init the anglerange
		$this->setAngleRange();

		// init the vertical align
		$this->setVerticalAlign();

		// init the align
		$this->setAlign();


		// init the anglerange
		$this->setBackground();
	}

	public function __destruct(){
		$this->width = 0;
		$this->height = 0;
	}

	/**
	 * Set the font family and size
	 *
	 * @param int $size
	 * @param string $color
	 * @return void
	 */
	function setFont(int $size = 15, string $color = '#000000'): void{
		$this->font = self::DEFAULT_FONT;

		// set the font size
		$this->font['size'] = [
			'min' => $size - 3,
			'max' => $size + 3
		];

		// set the font colors
		$colors = explode(',', $color);
		$this->font['color'] = [];
		foreach($colors as $color){
			$this->font['color'][] = self::hex2rgb($color);
		}
	}

	/**
	 * Set the font family and size
	 *
	 * @param string $subset
	 * @param string $case
	 * @param string $skip
	 * @return void
	 */
	function setCharacterSubset(string $subset = 'alphanum', string $case = 'mix', string $skip = ''): void{
		switch(strtolower($subset)){
			case 'alpha':
			case 'a-z':
				switch($case){
					case 'upper':
						$this->charactersubset = [[65, 90]];
						break;
					case 'lower':
						$this->charactersubset = [[97, 122]];
						break;
					default:
						$this->charactersubset = [[65, 90], [97, 122]];
				}
				break;
			case 'num':
			case '0-9':
				$this->charactersubset = [[48, 57]];
				break;
			case 'alphanum':
			case 'a-z0-9':
			case '0-9a-z':
			default:
				switch($case){
					case 'upper':
						$this->charactersubset = [[48, 57], [65, 90]];
						break;
					case 'lower':
						$this->charactersubset = [[48, 57], [97, 122]];
						break;
					default:
						$this->charactersubset = [[48, 57], [65, 90], [97, 122]];
				}
				break;
		}
		$skips = explode(',', $skip);
		$this->skip = [];
		foreach($skips as $skip){
			$this->skip[] = ord($skip);
		}
	}

	/* end: setCharacterSubset */

	/**
	 * Set the range of the angles
	 *
	 * @param int $angle
	 * @return void
	 */
	function setAngleRange(int $angle = 0): void{//FIXME currently only int is allowed in database
		$this->angle = [
			'left' => (int) ((-1) * $angle),
			'right' => $angle
		];
	}

	/* setAngleRange */

	/**
	 * Set the styles
	 *
	 * @param string $style
	 * @param string $color
	 * @return void
	 */
	public function setStyle(string $style = '', string $color = '#cccccc', string $number = '10,20'): void{
		$this->style = self::DEFAULT_STYLE;

		$styles = explode(',', $style);
		if(!empty($styles)){
			if(in_array('strikeout', $styles, true)){
				$this->style['strikeout'] = true;
			}
			if(in_array('fullcircle', $styles, true)){
				$this->style['fullcircle'] = true;
			}
			if(in_array('outlinecircle', $styles, true)){
				$this->style['outlinecircle'] = true;
			}
			if(in_array('fullrectangle', $styles, true)){
				$this->style['fullrectangle'] = true;
			}
			if(in_array('outlinerectangle', $styles, true)){
				$this->style['outlinerectangle'] = true;
			}
			$colors = explode(',', $color);
			if(count($colors) > 1){
				foreach($colors as $color){
					$this->style['color'][] = self::hex2rgb($color);
				}
			} else {
				$this->style['color'][] = self::hex2rgb($color);
			}
			$numbers = explode(',', $number);
			if(count($numbers) > 1){
				if($numbers[0] < $numbers[1]){
					$this->style['number'] = [
						'min' => $numbers[0],
						'max' => $numbers[1]
					];
				} else {
					$this->style['number'] = [
						'min' => $numbers[1],
						'max' => $numbers[0]
					];
				}
			} else {
				$this->style['number'] = [
					'min' => $number,
					'max' => $number
				];
			}
		}
	}

	/* setStyle */

	/**
	 * Set the vertical position
	 *
	 * @param string $valign
	 * @return void
	 */
	function setVerticalAlign(string $valign = 'random'): void{
		switch(strtolower($valign)){
			case 'top':
				$this->valign = 'top';
				break;
			case 'bottom':
				$this->valign = 'bottom';
				break;
			case 'middle':
				$this->valign = 'middle';
				break;
			case 'random':
			default:
				$this->valign = 'random';
				break;
		}
	}

	/* setVerticalAlign */

	/**
	 * Set the position
	 *
	 * @param string $align
	 * @return void
	 */
	function setAlign(string $align = 'random'): void{
		switch(strtolower($align)){
			case 'center':
				$this->align = 'center';
				break;
			case 'left':
				$this->align = 'left';
				break;
			case 'right':
				$this->align = 'right';
				break;
			case 'random':
			default:
				$this->align = 'random';
				break;
		}
	}

	/* setAlign */

	/**
	 * Set the background color
	 *
	 * @param string $background
	 * @return void
	 */
	function setBackground(string $background = "#ffffff", bool $transparent = false): void{
		$this->background = self::hex2rgb($background);
		$this->transparent = $transparent;
	}

	/* setBackground */

	/**
	 * Converts a Hex-code to rgb values
	 *
	 * @param string $hex
	 * @return int[]
	 */
	private static function hex2rgb(string $hex): array{
		$hex = preg_replace('/[^a-fA-F0-9]/', '', $hex);
		if(strlen($hex) == 3){
			return [
				(int) hexdec($hex[0] . $hex[0]),
				(int) hexdec($hex[1] . $hex[1]),
				(int) hexdec($hex[2] . $hex[2])
			];
		}
		if(strlen($hex) == 6){
			return [
				(int) hexdec($hex[0] . $hex[1]),
				(int) hexdec($hex[2] . $hex[3]),
				(int) hexdec($hex[4] . $hex[5])
			];
		}
		return [0, 0, 0];
	}

	/**
	 * @return array{int,int,int}
	 */
	private function getFont(float $size, string $sign): array{
		// Angle
		$angle = rand($this->angle['left'], $this->angle['right']);

		// Coordinates
		$coords = imagettfbbox($size, $angle, $this->fontfile, $sign);
		$width = intval(abs(( (-1) * abs(min($coords[0], $coords[6])) ) + ( abs(max($coords[2], $coords[4])) )));
		$height = intval(abs(( (-1) * abs(min($coords[1], $coords[7])) ) + ( abs(max($coords[3], $coords[5])) )));
		return [$angle, $width, $height];
	}

	/**
	 * Displayes the captcha image
	 * @throws we_exception_base
	 * @return resource
	 */
	function get(&$code)/* :resource */{
		$image = imagecreatetruecolor($this->width, $this->height);
		if(!$image){
			throw new we_exception_base('unable to create image');
		}
		$bgcolor = imagecolorallocate($image, $this->background[0], $this->background[1], $this->background[2]) ?: 0;
		if($this->transparent){
			imagecolortransparent($image, $bgcolor);
		}

		if($this->style['fullcircle'] || $this->style['fullrectangle'] || $this->style['outlinecircle'] || $this->style['outlinerectangle']){
			$counter = rand(intval($this->style['number']['min']), intval($this->style['number']['max']));

			$random = [];
			if($this->style['outlinecircle']){
				$random[] = 'outlinecircle';
			}
			if($this->style['outlinerectangle']){
				$random[] = 'outlinerectangle';
			}
			if($this->style['fullcircle']){
				$random[] = 'fullcircle';
			}
			if($this->style['fullrectangle']){
				$random[] = 'fullrectangle';
			}

			for($i = 0; $i < $counter; $i++){
				$do = $random[rand(0, count($random) - 1)];
				switch($do){
					case 'fullcircle':
						$cx = rand(0, $this->width);
						$cy = rand(0, $this->height);
						$w = rand(0, (int) ($this->width / 2));
						$h = rand(0, (int) ($this->height / 2));
						$color = $this->style['color'][rand(0, count($this->style['color']) - 1)];

						imagefilledarc($image, $cx, $cy, $w, $h, 0, 360, imagecolorallocate($image, $color[0], $color[1], $color[1]) ?: 0, IMG_ARC_PIE);
						break;
					case 'fullrectangle':
						$x1 = rand(0, $this->width);
						$y1 = rand(0, $this->height);
						$x2 = rand(0, (int) ($this->width / 2));
						$y2 = rand(0, (int) ($this->height / 2));
						$color = $this->style['color'][rand(0, count($this->style['color']) - 1)];

						imagefilledrectangle($image, $x1, $y1, $x2, $y2, imagecolorallocate($image, $color[0], $color[1], $color[1]) ?: 0);
						break;
					case 'outlinecircle':
						$cx = rand(0, $this->width);
						$cy = rand(0, $this->height);
						$w = rand(0, (int) ($this->width / 2));
						$h = rand(0, (int) ($this->height / 2));
						$color = $this->style['color'][rand(0, count($this->style['color']) - 1)];

						imagearc($image, $cx, $cy, $w, $h, 0, 360, imagecolorallocate($image, $color[0], $color[1], $color[1]) ?: 0);
					case 'outlinerectangle':
						$x1 = rand(0, $this->width);
						$y1 = rand(0, $this->height);
						$x2 = rand(0, (int) ($this->width / 2));
						$y2 = rand(0, (int) ($this->height / 2));
						$color = $this->style['color'][rand(0, count($this->style['color']) - 1)];

						imagerectangle($image, $x1, $y1, $x2, $y2, imagecolorallocate($image, $color[0], $color[1], $color[1]) ?: 0);
				}
			}
		}

		$code = '';
		$sumwidth = $xpos = $ypos = 0;
		$signs = [];

		for($counter = 0; $counter < $this->textlength; $counter++){
			// Sign
			$j = 0;
			$idx1 = rand(0, count($this->charactersubset) - 1);
			$sign = rand($this->charactersubset[$idx1][0], $this->charactersubset[$idx1][1]);
			if($this->skip){
				while(in_array($sign, $this->skip) || $j < 100){
					$j++;
					$idx1 = rand(0, count($this->charactersubset) - 1);
					$sign = rand($this->charactersubset[$idx1][0], $this->charactersubset[$idx1][1]);
				}
			}
			$sign = chr($sign);

			// Size
			$size = rand($this->font['size']['min'], $this->font['size']['max']);

			// Color
			$color = $this->font['color'][rand(0, count($this->font['color']) - 1)];

			// Family
			list($angle, $width, $height) = $this->getFont(rand($this->font['size']['min'], $this->font['size']['max']), $sign);

			// Abstand X-Position
			if(($xpos + $width) <= $this->width){
				// Y-Position
				switch($this->valign){
					case 'top':
						$ypos = $height + 5;
						break;
					case 'bottom':
						$ypos = $this->height - 5;
						break;
					case 'middle':
						$ypos = (int) (($this->height / 2) + ($height / 2));
						break;
					case 'random':
						$ypos = (int) rand($this->height - 5, $height + 5);
				}

				$signs[] = [
					'size' => $size,
					'angle' => $angle,
					'xpos' => $xpos,
					'ypos' => $ypos,
					'color' => imagecolorallocate($image, (int) $color[0], (int) $color[1], (int) $color[2]),
					'sign' => $sign,
				];


				$space = rand(intval(round($this->font['size']['min'] / 2)), intval(round($this->font['size']['max'] / 2)));

				// X-Position
				$xpos += (int) round($space) + $width;

				$code .= $sign;

				$sumwidth += round($space) + $width;
			}
		}

		if($this->align === 'random'){
			$temp = ['left', 'right', 'center'];
			$this->align = $temp[rand(0, 2)];
		}
		$xoffset = 0;
		switch($this->align){
			case 'left':
				$xoffset = 5;
				break;
			case 'right':
				$xoffset = (int) ($this->width - $sumwidth - 5);
				break;
			case 'center':
				$xoffset = (int) (($this->width / 2) - ($sumwidth / 2));
		}

		foreach($signs as $cursign){
			imagettftext(
				$image, // Imageressource
		$cursign['size'], // Fontsize
		(float) $cursign['angle'], // Angle
		$xoffset + $cursign['xpos'], // X-Position
		$cursign['ypos'], // Y-Position
		(int) $cursign['color'], // Fontcolor
		$this->fontfile, // Font Family (File)
		$cursign['sign'] // Text
			);
		}

		if($this->style['strikeout']){
			// Y-Position
			$max = (int) (($this->height / 2) + 10);
			$min = (int) (($this->height / 2) - 10);
			$y1 = $y2 = rand($min, $max);

			// X-Position
			$x1 = 0;
			$x2 = $this->width;

			$color = $this->font['color'][rand(0, count($this->font['color']) - 1)];

			imageline($image, $x1, $y1, $x2, $y2, imagecolorallocate($image, (int) $color[0], (int) $color[1], (int) $color[2]) ?: 0);
		}

		return $image;
	}

}
