<?php
declare(strict_types=1);

/**
 * Zend Framework
 *
 * LICENSE
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://framework.zend.com/license/new-bsd
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@zend.com so we can send you a copy immediately.
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

/**
 * Protocol_Abstract
 *
 * Provides low-level methods for concrete adapters to communicate with a remote mail server and track requests and responses.
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 * @version    $Id$
 * @todo Implement proxy settings
 */
abstract class we_mail_ProtocolAbstract{
	/** Mail default EOL string */
	protected const EOL = "\r\n";

	/** Default timeout in seconds for initiating session */
	const TIMEOUT_CONNECTION = 30;

	/** @var int Maximum of the transaction log */
	protected $_maximumLog = 64;

	/** @var string Hostname or IP address of remote server */
	protected $_host;

	/** @var int Port number of connection */
	protected $_port;

	/** @var resource Socket connection resource */
	protected $_socket;

	/** @var string Last request sent to server */
	protected $_request;

	/** @var array Array of server responses to last request */
	protected $_response;

	/** @var array Log of mail requests and server responses for a session */
	private $_log = [];

	/**
	 * Constructor.
	 *
	 * @param  string  $host OPTIONAL Hostname of remote connection (default: 127.0.0.1)
	 * @param  integer $port OPTIONAL Port number (default: null)
	 * @return void
	 */
	public function __construct(string $host = '127.0.0.1', int $port = 0){
		$this->_host = $host;
		$this->_port = $port;
	}

	/**
	 * Class destructor to cleanup open resources
	 *
	 * @return void
	 */
	public function __destruct(){
		$this->_disconnect();
	}

	/**
	 * Set the maximum log size
	 *
	 * @param integer $maximumLog Maximum log size
	 * @return void
	 */
	public function setMaximumLog(int $maximumLog): void{
		$this->_maximumLog = $maximumLog;
	}

	/**
	 * Get the maximum log size
	 *
	 * @return int the maximum log size
	 */
	public function getMaximumLog(): int{
		return $this->_maximumLog;
	}

	/**
	 * Create a connection to the remote host
	 *
	 * Concrete adapters for this class will implement their own unique connect scripts, using the _connect() method to create the socket resource.
	 */
	abstract public function connect();

	/**
	 * Retrieve the last client request
	 *
	 * @return string
	 */
	public function getRequest(): string{
		return $this->_request;
	}

	/**
	 * Retrieve the last server response
	 *
	 * @return array
	 */
	public function getResponse(): array{
		return $this->_response;
	}

	/**
	 * Retrieve the transaction log
	 *
	 * @return string
	 */
	public function getLog(): string{
		return implode('', $this->_log);
	}

	/**
	 * Reset the transaction log
	 *
	 * @return void
	 */
	public function resetLog(): void{
		$this->_log = [];
	}

	/**
	 * Add the transaction log
	 *
	 * @param  string $value new transaction
	 * @return void
	 */
	protected function _addLog(string $value): void{
		if($this->_maximumLog >= 0 && count($this->_log) >= $this->_maximumLog){
			array_shift($this->_log);
		}

		$this->_log[] = $value;
	}

	/**
	 * Connect to the server using the supplied transport and target
	 *
	 * An example $remote string may be 'tcp://mail.example.com:25' or 'ssh://hostname.com:2222'
	 *
	 * @param  string $remote Remote
	 * @throws we_exception_mail
	 * @return boolean
	 */
	protected function _connect(string $remote): bool{
		$errorNum = 0;
		$errorStr = '';

		// open connection
		$this->_socket = @stream_socket_client($remote, $errorNum, $errorStr, self::TIMEOUT_CONNECTION);

		if($this->_socket === false){
			if($errorNum == 0){
				$errorStr = 'Could not open socket';
			}
			throw new we_exception_mail($errorStr);
		}

		if(($result = $this->_setStreamTimeout(self::TIMEOUT_CONNECTION)) === false){
			throw new we_exception_mail('Could not set stream timeout');
		}

		return $result;
	}

	/**
	 * Disconnect from remote host and free resource
	 *
	 * @return void
	 */
	protected function _disconnect(): void{
		if(is_resource($this->_socket)){
			fclose($this->_socket);
		}
	}

	/**
	 * Send the given request followed by a LINEEND to the server.
	 *
	 * @param  string $request
	 * @throws we_exception_mail
	 * @return integer|boolean Number of bytes written to remote host
	 */
	protected function _send(string $request): int{
		if(!is_resource($this->_socket)){
			throw new we_exception_mail('No connection has been established to ' . $this->_host);
		}

		$this->_request = $request;

		$result = fwrite($this->_socket, $request . self::EOL);

		// Save request to internal log
		$this->_addLog($request . self::EOL);

		if($result === false){
			throw new we_exception_mail('Could not send request to ' . $this->_host);
		}

		return $result;
	}

	/**
	 * Get a line from the stream.
	 *
	 * @var    integer $timeout Per-request timeout value if applicable
	 * @throws we_exception_mail
	 * @return string
	 */
	protected function _receive(int $timeout = 0): string{
		if(!is_resource($this->_socket)){
			throw new we_exception_mail('No connection has been established to ' . $this->_host);
		}

		// Adapters may wish to supply per-commend timeouts according to appropriate RFC
		if($timeout !== 0){
			$this->_setStreamTimeout($timeout);
		}

		// Retrieve response
		$reponse = fgets($this->_socket, 1024) ?: '';

		// Save request to internal log
		$this->_addLog($reponse);

		// Check meta data to ensure connection is still valid
		$info = stream_get_meta_data($this->_socket);

		if(!empty($info['timed_out'])){
			throw new we_exception_mail($this->_host . ' has timed out');
		}

		if($reponse === false){
			throw new we_exception_mail('Could not read from ' . $this->_host);
		}

		return $reponse;
	}

	/**
	 * Parse server response for successful codes
	 *
	 * Read the response from the stream and check for expected return code.
	 * Throws a Mail_Protocol_Exception if an unexpected code is returned.
	 *
	 * @param  string|string[] $code One or more codes that indicate a successful response
	 * @throws we_exception_mail
	 * @return string Last line of response string
	 */
	protected function _expect($code, int $timeout = 0){
		$this->_response = [];
		$cmd = '';
		$msg = '';
		$errMsg = '';

		if(!is_array($code)){
			$code = [$code];
		}

		do{
			$this->_response[] = $result = $this->_receive($timeout);
			list($cmd, $more, $msg) = preg_split('/([\s-]+)/', $result, 2, PREG_SPLIT_DELIM_CAPTURE);

			if($errMsg !== ''){
				$errMsg .= ' ' . $msg;
			} elseif($cmd === null || !in_array($cmd, $code)){
				$errMsg = $msg;
			}
		} while(strpos($more, '-') === 0); // The '-' message prefix indicates an information string instead of a response string.

		if($errMsg !== ''){
			throw new we_exception_mail($errMsg . ': ' . $cmd);
		}

		return $msg;
	}

	/**
	 * Set stream timeout
	 *
	 * @param integer $timeout
	 * @return boolean
	 */
	protected function _setStreamTimeout(int $timeout): bool{
		return stream_set_timeout($this->_socket, $timeout);
	}

}
