<?php
declare(strict_types=1);

/**
 * Zend Framework
 *
 * LICENSE
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://framework.zend.com/license/new-bsd
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@zend.com so we can send you a copy immediately.
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

/**
 * Smtp implementation of Mail_Protocol_Abstract
 *
 * Minimum implementation according to RFC2821: EHLO, MAIL FROM, RCPT TO, DATA, RSET, NOOP, QUIT
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 */
class we_mail_ProtocolSmtp extends we_mail_ProtocolAbstract{
	/**
	 * The transport method for the socket
	 *
	 * @var string
	 */
	protected $_transport = 'tcp';

	/**
	 * Indicates that a session is requested to be secure
	 *
	 * @var string
	 */
	protected $_secure;

	/**
	 * Indicates an smtp session has been started by the HELO command
	 *
	 * @var bool
	 */
	protected $_sess = false;

	/**
	 * Indicates the HELO command has been issues
	 *
	 * @var bool
	 */
	protected $_helo = false;

	/**
	 * Indicates an smtp AUTH has been issued and authenticated
	 *
	 * @var bool
	 */
	protected $_auth = false;

	/**
	 * Indicates a MAIL command has been issued
	 *
	 * @var bool
	 */
	protected $_mail = false;

	/**
	 * Indicates one or more RCTP commands have been issued
	 *
	 * @var bool
	 */
	protected $_rcpt = false;

	/**
	 * Indicates that DATA has been issued and sent
	 *
	 * @var bool
	 */
	protected $_data = false;

	/**
	 * Constructor.
	 *
	 * @param  string  $host
	 * @param  integer $port
	 * @param  array   $config
	 * @return void
	 * @throws we_exception_mail
	 */
	public function __construct(string $host = '127.0.0.1', int $port = 0, array $config = []){
		if(isset($config['ssl'])){
			switch(strtolower($config['ssl'])){
				case 'tls':
					$this->_secure = 'tls';
					break;

				case 'ssl':
					$this->_transport = 'ssl';
					$this->_secure = 'ssl';
					$port = $port ?: 465;
					break;

				default:
					throw new we_exception_mail($config['ssl'] . ' is unsupported SSL type');
			}
		}

		// If no port has been specified then check the master PHP ini file. Defaults to 25 if the ini setting is null.
		$port = $port ?: ((int) ini_get('smtp_port')) ?: 25;
		parent::__construct($host, $port);
	}

	/**
	 * Connect to the server with the parameters given in the constructor.
	 *
	 * @return boolean
	 */
	public function connect(): bool{
		return $this->_connect($this->_transport . '://' . $this->_host . ':' . $this->_port);
	}

	/**
	 * Initiate HELO/EHLO sequence and set flag to indicate valid smtp session
	 *
	 * @param  string $host The client hostname or IP address (default: 127.0.0.1)
	 * @throws we_exception_mail
	 * @return void
	 */
	public function helo(string $host = '127.0.0.1'): void{
		// Respect RFC 2821 and disallow HELO attempts if session is already initiated.
		if($this->_sess === true){
			throw new we_exception_mail('Cannot issue HELO to existing session');
		}

		// Initiate helo sequence
		$this->_expect('220', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		$this->_ehlo($host);

		// If a TLS session is required, commence negotiation
		if($this->_secure == 'tls'){
			$this->_send('STARTTLS');
			$this->_expect('220', 180);
			if(!stream_socket_enable_crypto($this->_socket, true, STREAM_CRYPTO_METHOD_SSLv23_CLIENT)){
				throw new we_exception_mail('Unable to connect via TLS');
			}
			$this->_ehlo($host);
		}

		$this->_startSession();
		$this->auth();
	}

	/**
	 * Send EHLO or HELO depending on capabilities of smtp host
	 *
	 * @param  string $host The client hostname or IP address (default: 127.0.0.1)
	 * @throws we_exception_mail
	 * @return void
	 */
	protected function _ehlo(string $host): void{
		// Support for older, less-compliant remote servers. Tries multiple attempts of EHLO or HELO.
		try{
			$this->_send('EHLO ' . $host);
			$this->_expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		} catch (we_exception_mail $e){
			$this->_send('HELO ' . $host);
			$this->_expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		}
	}

	/**
	 * Issues MAIL command
	 *
	 * @param  string $from Sender mailbox
	 * @throws we_exception_mail
	 * @return void
	 */
	public function mail(string $from): void{
		if($this->_sess !== true){
			throw new we_exception_mail('A valid session has not been started');
		}

		$this->_send('MAIL FROM:<' . $from . '>');
		$this->_expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		// Set mail to true, clear recipients and any existing data flags as per 4.1.1.2 of RFC 2821
		$this->_mail = true;
		$this->_rcpt = false;
		$this->_data = false;
	}

	/**
	 * Issues RCPT command
	 *
	 * @param  string $to Receiver(s) mailbox
	 * @throws we_exception_mail
	 * @return void
	 */
	public function rcpt(string $to): void{
		if($this->_mail !== true){
			throw new we_exception_mail('No sender reverse path has been supplied');
		}

		// Set rcpt to true, as per 4.1.1.3 of RFC 2821
		$this->_send('RCPT TO:<' . $to . '>');
		$this->_expect(['250', '251'], 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		$this->_rcpt = true;
	}

	/**
	 * Issues DATA command
	 *
	 * @param  string $data
	 * @throws we_exception_mail
	 * @return void
	 */
	public function data($data): void{
		// Ensure recipients have been set
		if($this->_rcpt !== true){
			throw new we_exception_mail('No recipient forward path has been supplied');
		}

		$this->_send('DATA');
		$this->_expect('354', 120); // Timeout set for 2 minutes as per RFC 2821 4.5.3.2

		foreach(explode(we_mail_mime::LINEEND, $data) as $line){
			if(strpos($line, '.') === 0){
				// Escape lines prefixed with a '.'
				$line = '.' . $line;
			}
			$this->_send($line);
		}

		$this->_send('.');
		$this->_expect('250', 600); // Timeout set for 10 minutes as per RFC 2821 4.5.3.2
		$this->_data = true;
	}

	/**
	 * Issues the RSET command and validates answer
	 *
	 * Can be used to restore a clean smtp communication state when a transaction has been cancelled or commencing a new transaction.
	 *
	 * @return void
	 */
	public function rset(): void{
		$this->_send('RSET');
		// MS ESMTP doesn't follow RFC, see [ZF-1377]
		$this->_expect(['250', '220']);

		$this->_mail = false;
		$this->_rcpt = false;
		$this->_data = false;
	}

	/**
	 * Issues the NOOP command and validates answer
	 *
	 * Not used by Mail, could be used to keep a connection alive or check if it is still open.
	 *
	 * @return void
	 */
	public function noop(): void{
		$this->_send('NOOP');
		$this->_expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
	}

	/**
	 * Issues the VRFY command and validates answer
	 *
	 * Not used by Mail.
	 *
	 * @param  string $user User Name or eMail to verify
	 * @return void
	 */
	public function vrfy(string $user): void{
		$this->_send('VRFY ' . $user);
		$this->_expect(['250', '251', '252'], 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
	}

	/**
	 * Issues the QUIT command and clears the current session
	 *
	 * @return void
	 */
	public function quit(): void{
		if($this->_sess){
			$this->_send('QUIT');
			$this->_expect('221', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
			$this->_stopSession();
		}
	}

	/**
	 * Default authentication method
	 *
	 * This default method is implemented by AUTH adapters to properly authenticate to a remote host.
	 *
	 * @throws we_exception_mail
	 * @return void
	 */
	public function auth(): void{
		if($this->_auth === true){
			throw new we_exception_mail('Already authenticated for this session');
		}
	}

	/**
	 * Closes connection
	 *
	 * @return void
	 */
	public function disconnect(): void{
		$this->_disconnect();
	}

	/**
	 * Start mail session
	 *
	 * @return void
	 */
	protected function _startSession(): void{
		$this->_sess = true;
	}

	/**
	 * Stop mail session
	 *
	 * @return void
	 */
	protected function _stopSession(): void{
		$this->_sess = false;
	}

}
